/*
** GuessTheWord!
** Target: ATMEGA :: AVR-GCC
**
** Written by Caciula Raluca - 334CB
**
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lcd.h"
#include "dictionary.h"
#include "GuessTheWord.h"

#define BUFF_X_MAX      (LCD_X_RES)
#define BUFF_Y_MAX      (LCD_Y_RES / 8)

static uint8_t score;                              /* scorul jocului */
static uint8_t no_of_lives;                        /* numarul de vieti ramase */
static State state;                                /* starea in care se afla jocul */
static uint8_t bufferLCD[BUFF_Y_MAX][BUFF_X_MAX];  /* buffer pentru pixelii LCD-ului */
static Bool history[26];                           /* istoricul literelor introduse */
static uint8_t pos = -1;                           /* pozitia cuvantului de ghicit */
static uint8_t len_word;                           /* lungimea cuvantului de ghicit */
static uint8_t current_letter_pos;                 /* pozitia literei selectate de catre jucator */

/* programul principal */
int main(void)
{
    /* initializam LCD-ul */
    lcd_init();
    lcd_contrast(0x40);

    /* afisam textul de start si initializam jocul */
    print_start_text();
    init_game();
    _delay_ms(1000);

    
    /* in functie de starea jocului avem mai multe cazuri */
    while (TRUE) {
        switch (state) {
        case START:
            if (pressed_button(Button_ENTER)) {
                reset();
                state = PLAY_GAME;
            }

            break;
        case PLAY_GAME:
            if (pressed_button(Button_ENTER))
                clear_screen();
			play_game();
            _delay_ms(1000);
			
            break;
        case NEW_WORD:
            if (pressed_button(Button_ENTER)) {
				clear_screen();
                reset();
                state = PLAY_GAME;
                _delay_ms(1000);
            }

            break;
        default:
            return 1;
        }
    }

    return 0;
}

/*
 * Functie in care verificam starea butonului primit ca parametru.
 */
Bool pressed_button(Button button)
{
    return (PINB & (1 << button)) == 0 ? TRUE : FALSE;
}

/*
 * Functie in care setam pixelul de pe pozitia (x, y) la culoarea primita ca parametru.
 */
void set_pixel(uint8_t x, uint8_t y, Color color)
{
    uint8_t aux = LCD_Y_RES - y - 1;
    uint8_t row = aux / 8;

    /* determinam valoarea culorii pixelului in functie de cea primita ca parametru */
    switch (color) {
    case BLACK:
		bufferLCD[row][x] |= (1 << (aux % 8));
        break;
    case WHITE:
		bufferLCD[row][x] &= ~(1 << (aux % 8));
        break;
    default: 
		return;
    }

    /* pozitionam pointerul LCD-ului la coordonata (x, y) */
    lcd_send(0x80 | x, LCD_CMD);    // coloana
    lcd_send(0x40 | row, LCD_CMD);  // linie

    /* setam pixelul */
    lcd_send(bufferLCD[row][x], LCD_DATA);
}

/*
 * Functie in care stergem intreg ecranul LCD-ului (seteaza toti pixelii la culoarea WHITE).
 */
void clear_screen(void)
{
    for (uint8_t x = 0; x < LCD_X_RES; x++)
        for (uint8_t y = 0; y < LCD_Y_RES; y++)
            set_pixel(x, y, WHITE);
}

/*
 * Afisam textul pentru sfarsitul jocului.
 */
void print_end_text(void)
{
    char buf[11];
    sprintf(buf, "Score: %03u", score);
    clear_screen();
    lcd_goto_xy(4, 1);
    lcd_str("THE END!");
    lcd_goto_xy(3, 3);
    lcd_str(buf);
}

/*
 * Afisam textul pentru inceputul jocului.
 */
void print_start_text(void)
{
    clear_screen();
    lcd_goto_xy(2, 1);
    lcd_str("GuessTheWord!");
    lcd_goto_xy(1, 3);
    lcd_str("Caciula Raluca");
    lcd_goto_xy(3, 5);
    lcd_str("Press ENTER");
    lcd_goto_xy(4, 6);
    lcd_str("to start.");
}

/*
 * Afisam textul pentru pierderea jocului.
 */
void print_over_text(void)
{
    char buf[11];
    sprintf(buf, "Score: %03u", score);
    clear_screen();
    lcd_goto_xy(3, 1);
    lcd_str("GAME OVER!");
    lcd_goto_xy(3, 3);
    lcd_str(buf);
    lcd_goto_xy(3, 5);
    lcd_str("Press ENTER");
    lcd_goto_xy(3, 6);
    lcd_str("to replay.");
}

/*
 * Afisam textul pentru castigarea unei runde a jocului.
 */
void print_win_text(void)
{
    char buf[11];
    sprintf(buf, "Score: %03u", score);
    clear_screen();
    lcd_goto_xy(3, 1);
    lcd_str("GOOD JOB!");
    lcd_goto_xy(3, 3);
    lcd_str(buf);
    lcd_goto_xy((49 - len_word - 4) / 2, 4);
    lcd_str(to_guess_word[pos]);
    lcd_goto_xy(3, 5);
    lcd_str("Press ENTER");
    lcd_goto_xy(3, 6);
    lcd_str("to replay.");
}


/*
 * Afisam textul pentru scorul jocului.
 */
void draw_score_panel(void)
{
    lcd_goto_xy(10, 1);
    lcd_str("Score");
	char buf[4];
    lcd_goto_xy(11, 2);
    sprintf(buf, "%03u", score);
    lcd_str(buf);
	lcd_goto_xy(1, 1);
    lcd_str("*****");
    char buf2[4];
    lcd_goto_xy(2, 2);
    sprintf(buf2, "%03u", no_of_lives);
    lcd_str(buf2);
}

/*
 * Functie in care implementam functionalitatea jocului.
 */
void play_game(void)
{
    draw_score_panel();

    lcd_goto_xy((49 - len_word - 4) / 2, 3);
    lcd_str(to_guess_word[pos]);

	lcd_goto_xy(6, 6);
    lcd_chr('<');
	lcd_goto_xy(7, 6);
    lcd_chr(lowerLetters[current_letter_pos]);
	lcd_goto_xy(8, 6);
    lcd_chr('>');

	_delay_ms(1000);
    
    /* daca jucatorul a ghicit cuvantul */
    if (strcmp(to_guess_word[pos], words[pos]) == 0) {
        /* incrementam cu 10 scorul */
        score += 10;
        /* afisam textul corespunzator */
        print_win_text();

        /* aprindem LED-ul. */
        PORTC |= (1 << PC0);

        /* actualizam starea */
		state = NEW_WORD;

        _delay_ms(1000);
        return;
    }
        
    /* este apasat butonul LEFT */
    if (pressed_button(Button_LEFT)) {
        if (current_letter_pos - 1 < 0) {
            current_letter_pos = 25;
        } else {
            current_letter_pos--;
        }

    /* este apasat butonul RIGHT */
    } else if (pressed_button(Button_RIGHT)) {
        if (current_letter_pos + 1 > 25) {
            current_letter_pos = 0;
        } else {
            current_letter_pos++;
        }

    /* este apasat butonul UP */
    } else if (pressed_button(Button_UP) && history[current_letter_pos] == FALSE) {
        /* adaugam litera in istoric */
		history[current_letter_pos] = TRUE;

        /* daca litera introdusa este gresita */
		if (complete_word_with_letter() == FALSE) {
            /* decrementam numarul de vieti cu o unitate */
            no_of_lives--;

            /* daca numarul de vieti ajunge sa fie 0 jucatorul a pierdut */
            if (no_of_lives == 0) {
                /* resetam scorul */
				score = 0;

                /* afisam textul corespunzator */
                print_over_text();

                /*actualizam starea */
                state = NEW_WORD;

                _delay_ms(1000);
            }
       } else {
            /* daca litera introdusa este corecta incrementam scorul */
			score++;
	   }

       /* resetam pozitia literei alese de jucator */
	   current_letter_pos = 0;
    }
}

/*
 * Functie in care verificam daca litera introdusa se regaseste in cuvant.
 */
Bool complete_word_with_letter()
{
	Bool done = FALSE;

    if (strchr(words[pos], lowerLetters[current_letter_pos]) != NULL) {
        for (uint8_t i = 1; i < len_word - 1; i++) {
            if (words[pos][i] == lowerLetters[current_letter_pos] && to_guess_word[pos][i] == '_') {
                to_guess_word[pos][i] = lowerLetters[current_letter_pos];
				done = TRUE;
			}
        }
    }
    
    return done;
}

/*
 * Functie in care initializam jocul.
 */
void init_game(void)
{
    /* setam toti pinii portului C ca input si activam rezistentele de pull-up */
    DDRB = 0x00;
    PORTB = 0xFF;

    /* setam pinul PC0 ca iesire */
    DDRC |= (1 << PC0);

    /* stingem LED-ul */
    PORTC &= ~(1 << PC0);

    /* dezactivam portul JTAG */
    MCUCR = _BV(JTD);
    MCUCR = _BV(JTD);

    /* initializam numarul de incercari */
    no_of_lives = 5;

    /* initializam starea jocului */
    state = START;

    /* initializam scorul */
    score = 0;
}

/*
 * Functie in care resetam caracteristicile jocului.
 */
void reset(void) {
    /* resetam numarul de incercari */
    no_of_lives = 5;

    /* resetam istoricul literelor introduse */
    memset(history, FALSE, 26 * sizeof(Bool));
    
    /* trecem la urmatorul cuvant din dictionar */
    pos++;
    if (pos >= NO_MAX_WORDS) {
        print_end_text();

        _delay_ms(7000);
        exit(1);
    }
    
    /* determinam lungimea noului cuvant */
    len_word = strlen(words[pos]);

    /* resetam pozitia literei curente */
    current_letter_pos = 0;
    
    /* stingem LED-ul */
    PORTC &= ~(1 << PC0);
}
